#include <windows.h>
#include <stdio.h>
#include <ctype.h>
#include "dialog.h"
#include "pcresrch.h"
#include <pcre.h>

int mode; //MSG_CMD_PCRE_Search or MSG_CMD_PCRE_Replace
int within_block;
int all;
int last_line;
int last_pos;
int cur_line;
int delta=1;
int end_line; // Line at which to end search
char *cur_string=NULL;
pcre *pcre_pattern=NULL;
int pcre_options=0;
const char *pcre_error;
int pcre_errptr;
int substr_arr[100*3];
struct EditorInfo ei;
static struct PluginStartupInfo Info;

int Dialog(UnpackedDialog *d);
void report_regexp_error();
void position_cursor(int line,int pos,int mark,int sel_size);
int expand_replace_pattern(char *buf,char *pattern,char *match,int *asubstr,int num_substr);
HANDLE init_search_dialog();
HANDLE init_replace_dialog();

UnpackedDialog d_search={NULL},d_replace={NULL};
static char *mbox_not_found[]=
{ "Not found","Pattern not found","Ok" };
static char *mbox_regexp_err[]=
{ "Error in regexp",NULL,NULL,"Cancel" };
static char *mbox_replace[]=
{ "Match found",NULL,"matched, replace with",NULL,"&Replace","&All","&Skip","&Cancel" };
enum {REPLACE_CMD_REPLACE,REPLACE_CMD_ALL,REPLACE_CMD_SKIP,REPLACE_CMD_CANCEL};

void WINAPI _export SetStartupInfo(struct PluginStartupInfo *Info)
{
  ::Info=*Info;
}

DialogItem dialog_search[]=
{
  { DI_EDIT,0,1,53,0,TRUE,FALSE,0,0,""},
  { DI_TEXT,0,0,0,0,FALSE,FALSE,0,0,(char*)MSG_Search_for },
  { DI_CHECKBOX,0,2,0,0,FALSE,FALSE,0,0,(char*)MSG_Case_sensetive},
  { DI_CHECKBOX,0,3,0,0,FALSE,FALSE,0,0,(char*)MSG_Reverse_search},
  { DI_CHECKBOX,0,4,0,0,FALSE,FALSE,0,0,(char*)MSG_Within_block},
  { DI_CHECKBOX,0,5,0,0,FALSE,FALSE,0,0,(char*)MSG_Mark_found},
//  { DI_CHECKBOX,0,6,0,0,FALSE,FALSE,0,0,(char*)MSG_Interactive},
  { DI_TEXT,0,6,54,0,FALSE,FALSE,0,0,""},
  { DI_BUTTON,14,7,0,0,FALSE,FALSE,0,TRUE,(char*)MSG_BUT_Search},
  { DI_BUTTON,28,7,0,0,FALSE,FALSE,0,0,(char*)MSG_BUT_Cancel}
};

DialogItem dialog_replace[]=
{
  { DI_EDIT,0,1,53,0,TRUE,FALSE,0,0,""},
  { DI_EDIT,0,3,53,0,FALSE,FALSE,0,0,""},
  { DI_CHECKBOX,0,4,0,0,FALSE,FALSE,0,0,(char*)MSG_Case_sensetive},
  { DI_CHECKBOX,0,5,0,0,FALSE,FALSE,0,0,(char*)MSG_Reverse_search},
  { DI_TEXT/*CHECKBOX*/,0,6,0,0,FALSE,FALSE,0,0,(char*)MSG_Within_block},
  { DI_CHECKBOX,0,7,0,0,FALSE,FALSE,0,0,(char*)MSG_All_occurances},
  { DI_TEXT,0,8,54,0,FALSE,FALSE,0,0,""},
  { DI_BUTTON,14,9,0,0,FALSE,FALSE,0,TRUE,(char*)MSG_BUT_Replace},
  { DI_BUTTON,28,9,0,0,FALSE,FALSE,0,0,(char*)MSG_BUT_Cancel},
  { DI_TEXT,0,0,0,0,FALSE,FALSE,0,0,(char*)MSG_Search_for },
  { DI_TEXT,0,2,0,0,FALSE,FALSE,0,0,(char*)MSG_Replace_with },
};

enum {ITEM_PATTERN,ITEM_REPLACE_PATTERN,ITEM_CB_CASE,ITEM_CB_REVERSE,
ITEM_CB_WITHIN_BLOCK,ITEM_CB_MARK_FOUND,ITEM_CB_ALL=ITEM_CB_MARK_FOUND,ITEM_CB_INTERACTIVE};



void dumpstr(char *str)
{
  Info.EditorControl(ECTL_INSERTTEXT,str);
}

HANDLE WINAPI _export OpenPlugin(int OpenFrom,int Item)
{
  UnpackedDialog *d;
  EditorConvertText ect;
  all=0;

  if (Item!=MSG_CMD_Continue_Search)
  {
    Info.EditorControl(ECTL_GETINFO,&ei);

    mode=Item;

    if (Item==MSG_CMD_PCRE_Search)
    {
      if (init_search_dialog()==INVALID_HANDLE_VALUE)
        			return(INVALID_HANDLE_VALUE);
      d=&d_search;
    }
    else
    {
      if (init_replace_dialog()==INVALID_HANDLE_VALUE)
      				return(INVALID_HANDLE_VALUE);
      d=&d_replace;
    }

/*    if (Item==MSG_CMD_PCRE_Search)
    {
      d_search.d[ITEM_CB_WITHIN_BLOCK].Selected=(ei.BlockType==BTYPE_NONE)?0:1;
    }    */

Retry:
    int DialogCode=Dialog(d);
  
    if (DialogCode<0)    return(INVALID_HANDLE_VALUE);

    if (mode==MSG_CMD_PCRE_Search)
    	within_block=d_search.d[ITEM_CB_WITHIN_BLOCK].Selected;
    else
    	within_block=0;

/*  dumpstr("/");
  dumpstr(d[2].Data);
  dumpstr("/");
*/

    if (pcre_pattern) free(pcre_pattern);

/* If we have anchored pattern, i.e. one that will only
   match at bol, we'd like (later) to continue just from
   next line, when it matched (or else we'd get nonsencial
   extra matches). To achive this, we have to know that damned
   thing anchored. pcre_info supposed to do this, but it fakes.
   To workaround, we use PCRE_MULTILINE option */
                         
    pcre_options=/*0*/PCRE_MULTILINE;

    if (!d->d[ITEM_CB_CASE].Selected)
    			pcre_options|=PCRE_CASELESS;

    pcre_pattern=pcre_compile(d->d[ITEM_PATTERN].Data,
                              pcre_options,
                              &pcre_error,
                              &pcre_errptr,
                              NULL);

    if (pcre_pattern==NULL)
    {
      report_regexp_error();
      goto Retry;
    }

    delta=d->d[ITEM_CB_REVERSE].Selected?-1:1;

    if (mode==MSG_CMD_PCRE_Replace &&
        d_replace.d[ITEM_CB_ALL].Selected)    	all=1;

    if (!within_block || ei.BlockType==BTYPE_NONE)
    {
      cur_line=ei.CurLine; last_pos=ei.CurPos;
      end_line= (delta==1)?ei.TotalLines:-1;
    }
    else
    {
      if (delta==1)
      {
        cur_line=ei.BlockStartLine; last_pos=-1;
        end_line=-1;
      }
      else
      {
        struct EditorGetString egs;
        egs.StringNumber=ei.BlockStartLine-1;
        do
        {
          ++egs.StringNumber;
          if (!Info.EditorControl(ECTL_GETSTRING,&egs))
          				return(INVALID_HANDLE_VALUE);
        } while (egs.SelStart!=-1);
        cur_line=egs.StringNumber;
        end_line=ei.BlockStartLine-1;
      }
    }

  }
  else cur_line=last_line;

  {
    int pcre_st;
    int first=1;

/* Main search loop */

    while (1)
    {
      struct EditorGetString egs;
      egs.StringNumber=cur_line;

      /* We've done if... */

      if (!Info.EditorControl(ECTL_GETSTRING,&egs) || // FAR denies us a line
          cur_line==ei.TotalLines || 		      // We're past last line
          cur_line==end_line ||			      // We reached stop line
           (within_block &&
            ei.BlockType!=BTYPE_NONE &&
            (egs.SelStart==-1 || egs.SelStart==egs.SelEnd))) // or selection has ended
      {
        Info.Message(Info.ModuleNumber,
                     FMSG_WARNING,NULL,
                     mbox_not_found,
                     sizeofA(mbox_not_found),
                     1);
        break;
      }

/*      {
        char buf[50];
        sprintf(buf,"%d",egs.SelStart);
        dumpstr(buf);
        dumpstr("-");
        sprintf(buf,"%d",egs.SelEnd);
        dumpstr(buf);
      }
*/

/* if continue search, or started non-block search, we're matching
   not from the very beginning of line */

        /* line position magic */

      if (first)
      {
        if (Item==MSG_CMD_Continue_Search||ei.BlockType==BTYPE_NONE)
      				     egs.SelStart=last_pos+1; first=0;
      }
      else if (!within_block) egs.SelStart=0;

      {

        if (within_block)
        {
          if (egs.SelStart==-1) egs.SelStart=0;
          if (egs.SelEnd==0 || egs.SelEnd==-1 || egs.SelEnd>egs.StringLength)
            egs.SelEnd=egs.StringLength;
        }
        else { egs.SelEnd=egs.StringLength; } 

        /* let's get a copy of line ... */

        if (cur_string) free(cur_string);
        cur_string=(char *)malloc(egs.StringLength);
        memcpy(cur_string,egs.StringText,egs.StringLength);
  
        /* ... and convert it to console charset */

        ect.Text=cur_string+egs.SelStart;
        ect.TextLength=egs.SelEnd-egs.SelStart/*+1*/;
        Info.EditorControl(ECTL_EDITORTOOEM,&ect);
  
        /* ok, let's PCRE-match it! */

        pcre_st=pcre_exec(pcre_pattern,NULL,ect.Text,ect.TextLength,
                  0,0,substr_arr,sizeof(substr_arr));
/*        {
          char buf[50];
          sprintf(buf,"%d",pcre_st);
  //        dumpstr(buf);
        }*/
      }

      cur_line+=delta;

      /* if we haven't matched this line, move to other */

      if (pcre_st==PCRE_ERROR_NOMATCH) continue;

      /* ok, we've matched, pcre_st is number of subpatterns */

      if (pcre_st>=0)
      {
        /* match was here */

        last_line=cur_line-delta;
        last_pos=egs.SelStart+substr_arr[0];

        position_cursor(last_line,last_pos,
        		mode==MSG_CMD_PCRE_Replace || d_search.d[ITEM_CB_MARK_FOUND].Selected,
                        substr_arr[1]-substr_arr[0]);

        /* if pattern is anchored, next time we'll start at once
          from next line */

        {
          int opt,ch;
          if (pcre_info(pcre_pattern,&opt,&ch)>=0)
          {
            static char a[2]=" ",buf[10];
            if (ch=='^' || ch==-1) { last_pos=-1; last_line+=delta; }
          }
        }

        if (mode==MSG_CMD_PCRE_Replace)
        {
          static char buf1[50]="\"",buf2[50];
          static char buf[1000];
          char *p;

          /* expanding replacement, prepairing ask dialog,
             reconstructing whole line to replace */

          if (!all)
          {
            memcpy(buf1+1,
                   cur_string+egs.SelStart+substr_arr[0],
                   substr_arr[1]-substr_arr[0]);
            p=buf1+substr_arr[1]-substr_arr[0]+1;
            *p++='\"'; *p=0;
          }

          p=buf;
          memcpy(p,cur_string,egs.SelStart+substr_arr[0]);
          p+=egs.SelStart+substr_arr[0];
          int len=expand_replace_pattern(
                      p,
              	      d_replace.d[ITEM_REPLACE_PATTERN].Data,
                      ect.Text,
                      substr_arr,
                      pcre_st);
          ect.Text=p;
          ect.TextLength=len;
          if (!all) sprintf(buf2,"\"%.44s\"?",p);
          Info.EditorControl(ECTL_OEMTOEDITOR,&ect);
          p+=len;
          len=egs.StringLength-(egs.SelStart+substr_arr[1]);
          memcpy(p,
                 cur_string+egs.SelStart+substr_arr[1],
                 len);
          p+=len; *p=0;

          int code;
          if (all) code=REPLACE_CMD_REPLACE;
          else
          {
            mbox_replace[1]=buf1;
            mbox_replace[3]=buf2;
  
            Info.EditorControl(ECTL_REDRAW,NULL);
            code=Info.Message(Info.ModuleNumber,
                         FMSG_DOWN,NULL,
                         mbox_replace,
                         sizeofA(mbox_replace),
                         4);
          }

          if (code!=-1 && code!=REPLACE_CMD_CANCEL)
          {
            cur_line=last_line;
            first=1;
            if (code==REPLACE_CMD_SKIP) continue;

            EditorSetString ess;
            ess.StringNumber=-1;
            ess.StringText=buf;
            ess.StringEOL=NULL;
            ess.StringLength=p-buf;
            Info.EditorControl(ECTL_SETSTRING,&ess);

            if (code==REPLACE_CMD_ALL) all=1;

            continue;
          }

        }

        return(INVALID_HANDLE_VALUE);
      }
      break;
    }
  }
  return(INVALID_HANDLE_VALUE);
}

/* Set cursor, and if needed, mark blocky */

void position_cursor(int line,int pos,int mark,int sel_size)
{
  EditorSetPosition esp;
  EditorInfo ei;
  Info.EditorControl(ECTL_GETINFO,&ei);
  esp.CurLine=line;
  esp.CurPos=pos;
  esp.CurTabPos=-1;
  esp.Overtype=-1;
  if (line-ei.TopScreenLine>6) esp.TopScreenLine=line-3;
  else esp.TopScreenLine=-1;
  esp.LeftPos=-1;
  Info.EditorControl(ECTL_SETPOSITION,&esp);
  if (mark)
  {
    EditorSelect es;
    es.BlockType=BTYPE_STREAM;
    es.BlockStartLine=line;
    es.BlockStartPos=pos;
    es.BlockWidth=sel_size;
    es.BlockHeight=1;
    Info.EditorControl(ECTL_SELECT,&es);
  }
}

int expand_replace_pattern(char *buf,char *pattern,char *match,int *asubstr,int num_substr)
{
  char *p=buf;

  while (*pattern)
  {
    if (*pattern=='\\')
    {
      int num=0;
      if (pattern[1]=='\\') { *p++='\\'; pattern+=2; continue; }
      else if (pattern[1]=='&') pattern+=2;
      else if (isdigit(pattern[1]))
      {
        ++pattern;
        while (isdigit(*pattern))
        {
          num*=10; num+=*pattern++&0xf;
        }
        if (num>=num_substr) num=num_substr-1;
      }
      else goto Usual;

      int len=asubstr[num*2+1]-asubstr[num*2];
      memcpy(p,match+asubstr[num*2],len);
      p+=len;
      continue;
    }

Usual:
  *p++=*pattern++;
  }
  *p=0;
  return p-buf;
}

void WINAPI _export GetPluginInfo(struct PluginInfo *Info)
{
  Info->StructSize=sizeof(*Info);
  Info->Flags=PF_EDITOR|PF_VIEWER|PF_DISABLEPANELS;
  Info->DiskMenuStringsNumber=0;
  static char *PluginMenuStrings[LAST_MSG_CMD];
  PluginMenuStrings[MSG_CMD_PCRE_Search]=GetMsg(MSG_CMD_PCRE_Search);
  PluginMenuStrings[MSG_CMD_PCRE_Replace]=GetMsg(MSG_CMD_PCRE_Replace);
  PluginMenuStrings[MSG_CMD_Continue_Search]=GetMsg(MSG_CMD_Continue_Search);
  Info->PluginMenuStrings=PluginMenuStrings;
  Info->PluginMenuStringsNumber=sizeofA(PluginMenuStrings);
  Info->PluginConfigStringsNumber=0;

  mbox_regexp_err[0]=GetMsg(MSG_TITLE_Error_in_regexp);
  mbox_regexp_err[3]=GetMsg(MSG_BUT_Cancel);

  mbox_not_found[0]=GetMsg(MSG_TITLE_Not_found);
  mbox_not_found[1]=GetMsg(MSG_Not_found);
  mbox_not_found[2]=GetMsg(MSG_BUT_Ok);

  mbox_replace[0]=GetMsg(MSG_TITLE_Match_found);
  mbox_replace[2]=GetMsg(MSG_matched_replace_with);
  mbox_replace[4]=GetMsg(MSG_BUT_Replace);
  mbox_replace[5]=GetMsg(MSG_BUT_All);
  mbox_replace[6]=GetMsg(MSG_BUT_Skip);
  mbox_replace[7]=GetMsg(MSG_BUT_Cancel);
}

void report_regexp_error()
{
  static char msg_txt[90];
  sprintf(msg_txt,GetMsg(MSG_Error_in_regexp),pcre_errptr);
  mbox_regexp_err[1]=msg_txt;
  mbox_regexp_err[2]=(char*)(pcre_error?pcre_error:"(null)");
  Info.Message(Info.ModuleNumber,FMSG_WARNING,NULL,mbox_regexp_err,sizeofA(mbox_regexp_err),1);
}

HANDLE init_search_dialog()
{
    if (!d_search.d)
    {
      if(!MakeFarDialog((char*)MSG_TITLE_PCRE_Search,dialog_search,
    					       sizeofA(dialog_search),
    					       &d_search))
    		  return(INVALID_HANDLE_VALUE);
    }
}

HANDLE init_replace_dialog()
{
    if (!d_replace.d)
    {
      if(!MakeFarDialog((char*)MSG_TITLE_PCRE_Replace,dialog_replace,
    					       sizeofA(dialog_replace),
    					       &d_replace))
    		  return(INVALID_HANDLE_VALUE);
    }
}



char *GetMsg(int MsgId)
{
  return(Info.GetMsg(Info.ModuleNumber,MsgId));
}

int Dialog(UnpackedDialog *d)
{
  return Info.Dialog(Info.ModuleNumber,-1,-1,
    	             d->xsize,d->ysize,
    	             "Contents",d->d,d->num);
}
